/*
    Defines ri-graph directive
    - display nodes
    - handle ground click
*/
angular.module('ri.graph', [])

.directive('riGraph', [function() {
    return {
        restrict: 'EA',
        scope: {
            'nodes' : '=',
            'bombs' : '=',
            'flags' : '=',
            'onclickGround' : '&',
            'onrightclickGround' : '&',
            'onrightclickFlag' : '&',
            'onrightclickProbe' : '&',
            'onclickProbe' : '&',
            'api' : '=',
            'models' : '='
        },
        link: function(scope, element, attrs) {

            /*
             * INIT
             */

            var zoom = d3.behavior.zoom()
                .scaleExtent([1, 3])
                .on("zoom", zoomed);

            var groundSize = 500, margin = 10, viewBoxSize = groundSize+2*margin;
            var svg = d3.select(element[0])
                        .append("svg")
                            .attr("viewBox", (-margin) + " " + (-margin) + " " +(viewBoxSize)+" "+(viewBoxSize))
                            .style({
                                'width' : '100%'
                            })
            var container =    svg
                    .append("g")
                        .call(zoom)
                    .append("g");

            var gridpattern = '<pattern id="pattern" width="25" height="25" patternUnits="userSpaceOnUse"><path d="M 0 0 L 25 0 25 25 0 25 z" stroke="#222" stroke-width="1" fill="none"></path></pattern>';
            svg.append('defs').html(gridpattern);

            container.append('rect').attr({
                x:-margin, y:-margin, width:viewBoxSize, height:viewBoxSize
            }).style({
                //'fill-opacity' : 0,
                'fill' : 'url(#pattern)',
                'stroke' : "#222",
                'stroke-width' : 4
            });
            var areaLayer = container.append('g');
            var clearedareaLayer = container.append('g');
            var itemLayer = container.append('g');
            var nodeLayer = container.append('g');
            var bombLayer = container.append('g');
            var ghost = svg.append('g').style('opacity', 0);

            var ghostCursor = ghost.append('g').style({'stroke-width':2, 'stroke' : '#eee', 'fill':'none'});
            ghostCursor.append('line').attr({x1 : -10, y1 : 0, x2 : -6, y2 : 0})
            ghostCursor.append('line').attr({x1 : 10, y1 : 0, x2 : 6, y2 : 0})
            ghostCursor.append('line').attr({x1 : 0, y1 : -10, x2 : 0, y2 : -6})
            ghostCursor.append('line').attr({x1 : 0, y1 : 10, x2 : 0, y2 : 6})
            ghostCursor.append('circle').attr('r', 9);
            var ghostArea = ghost.append('circle').attr('r', scope.models.find(function(m) {return m.type=='probe'}).area).style({
                'fill' : 'none',
                'stroke' : '#eee',
                'stroke-dasharray' : 4
            });


            // Zoom behavior with bounded pan. Thanks http://bl.ocks.org/shawnbot/6518285
            function zoomed() {
              // the "zoom" event populates d3.event with an object that has
              // a "translate" property (a 2-element Array in the form [x, y])
              // and a numeric "scale" property
              var e = d3.event,
                  // now, constrain the x and y components of the translation by the
                  // dimensions of the viewport
                  tx = Math.min(0, Math.max(e.translate[0], viewBoxSize - viewBoxSize * e.scale)),
                  ty = Math.min(0, Math.max(e.translate[1], viewBoxSize - viewBoxSize * e.scale));
              // then, update the zoom behavior's internal translation, so that
              // it knows how to properly manipulate it on the next movement
              zoom.translate([tx, ty]);
              // and finally, update the <g> element's transform attribute with the
              // correct translation and scale (in reverse order)
              container.attr("transform", "translate(" + [tx, ty] + ")scale(" + d3.event.scale + ")");
              ghostArea.attr("transform", "scale(" + d3.event.scale + ")");
              //container.attr("transform", "scale(" + d3.event.scale + ")");
            }

            /*
             * INPUTS
             */

            container.on('click', function() {
                var point = d3.mouse(this),
                    p = {x: point[0], y: point[1] };
                scope.onclickGround({pos:p});
            })
            container.on("contextmenu", function (d, i) {
                d3.event.preventDefault();
                // react on right-clicking
                var point = d3.mouse(this),
                    p = {x: point[0], y: point[1] };
                scope.onrightclickGround({pos:p});
            });
            svg.on("mousemove", function() {
                /*console.log(this);
                console.log(d3.mouse(this));*/
                ghost.attr('transform', 'translate('+d3.mouse(this)+')');
            })
            svg.on('mouseenter', function() {
                ghost.style('opacity', 1);
                ghostArea.style('opacity', 1);
                ghost.attr('transform', 'translate('+d3.mouse(this)+')');
            })
            svg.on('mouseleave', function() {
                ghost.style('opacity', 0);
            })


            nodeLayer.on('mouseover', function() {
                ghostArea.style('opacity', 0);
            });
            nodeLayer.on('mouseout', function() {
                ghostArea.style('opacity', 1);
            });

            /*
             * RENDERING
             */

            // watch for data changes and re-render
            scope.$watch('nodes', function(newVals, oldVals) {
                scope.render();
            }, true);
            scope.$watch('flags', function(newVals, oldVals) {
                scope.render();
            }, true);
            scope.$watch('bombs', function(newVals, oldVals) {
                scope.render();
            }, true);

            scope.render = function() {
                /* PROBES */
                var probes = nodeLayer.selectAll('.node').data(scope.nodes, function(n) {return n.id});
                var newprobes = probes.enter()
                        .append('g')
                            .attr({
                                class : 'node',
                                transform : function(d) {return 'translate('+ [d.pos.x, d.pos.y] +')'}
                            });
                newprobes.append('circle')
                    .attr({
                        class : 'center'
                    })
                    .style({
                        //fill : function(d) {return d.out}
                        fill : '#222'
                    })
                    .on('mouseenter', function(probe) {
                        if (!probe.cleared) {
                            d3.select(this).transition().attr('r', function(d) {return d.size+3});
                        }
                    })
                    .on('mouseleave', function(probe) {
                        if (!probe.cleared) {
                            d3.select(this).transition().attr('r', function(d) {return d.size});
                        }
                    })
                    .on('click', function(d) {
                        d3.event.preventDefault();
                        d3.event.stopPropagation();
                        var valid = scope.onclickProbe({probe:d});
                        if (valid === false) {
                            d3.select(this.parentNode).select('.area-outline')
                                .style('fill-opacity', 0.5)
                                .transition().duration(1000).style('fill-opacity', 0);
                        }

                    })
                    .on('contextmenu', function(d) {
                        d3.event.preventDefault();
                        d3.event.stopPropagation();
                        scope.onrightclickProbe({probe:d});
                    });
                newprobes.append('text')
                    //.text(function(d) {return d.items.length})
                    .attr('dy', '.3em')
                    .style({
                        'text-anchor' : 'middle',
                        'pointer-events' : 'none',
                        'fill' : '#ddd',
                        'stroke' : '#ddd',
                        'stroke-width' : 1
                    })
                newprobes.append('circle')
                    .attr({
                        class : 'area-outline',
                        r : function(d) {return d.area}
                    })
                    .style({
                        'stroke' : '#222',
                        'stroke-dasharray' : 2,
                        'fill' : 'red',
                        'fill-opacity' : 0,
                        'pointer-events' : 'none'
                    });
                probes.select('.center').attr('r', function(d) {return d.cleared?3:d.size})
                probes.select('text').text(function(d) {return d.disabled?'x':(d.cleared?'':d.items.length)})
                probes.exit().remove();

                var itemGroups = itemLayer.selectAll('.item-group').data(scope.nodes, function(n) {return n.id});
                itemGroups.enter()
                    .append('g')
                        .attr({
                            class : 'item-group',
                            transform : function(d) {return 'translate('+ [d.pos.x, d.pos.y] +')'}
                        });
                itemGroups.style('opacity', function(d) {return (d.cleared||d.disabled) ? 0 : 1});
                itemGroups.exit().remove();

                var items = itemGroups.selectAll('.item').data(function(d) {return d.items})
                items.enter()
                        .append('circle')
                            .attr({
                                class : 'item',
                                //'vector-effect' : 'non-scaling-stroke'
                            })
                            .style({
                                'fill' : 'none',
                                'stroke' : '#ddd',
                                'stroke-width' : 2
                            });
                items//.transition().duration(500)
                    .attr('r', function(d) {return d.dist});
                items.exit().remove();

                var probesArea = areaLayer.selectAll('.probe-area').data(scope.nodes, function(n) {return n.id});
                probesArea.enter()
                    .append('g')
                        .attr({
                            class : 'probe-area',
                            transform : function(d) {return 'translate('+ [d.pos.x, d.pos.y] +')'}
                        })
                        .append('circle')
                            .attr({
                                r : function(d) {return d.area}
                            })
                            .style({
                                /*'fill' : 'white',
                                'opacity' : 0.2*/
                                'fill' : '#336E7B',
                            });
                probesArea.style('opacity', function(d) {return d.disabled ? 0 : 1});
                probesArea.exit().remove();

                var clearedArea = clearedareaLayer.selectAll('.cleared-area').data(scope.nodes, function(n) {return n.id});
                clearedArea.enter()
                    .append('g')
                        .attr({
                            class : 'cleared-area',
                            transform : function(d) {return 'translate('+ [d.pos.x, d.pos.y] +')'}
                        })
                        .append('circle')
                            .attr({
                                r : function(d) {return d.area}
                            })

                clearedArea.style('fill', function(d) {return d.cleared ? '#59ABE3' : 'none'})
                clearedArea.style('opacity', function(d) {return d.disabled ? 0 : 1});
                clearedArea.exit().remove();

                /* FLAGS */
                var flags = nodeLayer.selectAll('.flag').data(scope.flags)
                flags.enter()
                    .append('circle')
                        .attr({
                            class : 'flag'
                        })
                        .style({
                            fill : 'darkred',
                            stroke : '#222',
                            'stroke-width' : 2
                        })
                        .on('click', function(d) {
                            d3.event.stopPropagation();
                        })
                        .on('contextmenu', function(d) {
                            d3.event.preventDefault();
                            d3.event.stopPropagation();
                            scope.onrightclickFlag({flag:d});
                        });
                flags.attr({
                    cx : function(d) {return d.pos.x},
                    cy : function(d) {return d.pos.y},
                    r : function(d) {return d.area}
                })
                flags.exit().remove();

                /* BOMBS */
                var bombs = bombLayer.selectAll('.bomb')
                                //.data(scope.bombs)
                                .data(
                                    scope.bombs.filter(function(b) {return b.exploded || b.disabled}),
                                    function(b) {return b.id}
                                );
                bombs.enter()
                    .append('circle')
                        .attr({
                            class : 'bomb',
                            r : 0
                        })
                        .style({
                            stroke : '#222',
                            'stroke-width' : 2
                        })
                        .transition().duration(function(d){return d.exploded?500:0}).ease('elastic')
                        .attr('r', 8);
                bombs.attr({
                        cx : function(d) {return d.pos.x},
                        cy : function(d) {return d.pos.y},
                    })
                    .style({
                        fill : function(d) {return d.disabled ? 'green' : 'red'},
                    })
            }

            scope.api.showBombs = function() {
                bombLayer.selectAll('.bomb').data(scope.bombs,function(b) {return b.id})
                    .enter()
                        .append('circle')
                            .attr({
                                class : 'bomb',
                                cx : function(d) {return d.pos.x},
                                cy : function(d) {return d.pos.y},
                                r : 8
                            })
                            .style({
                                fill : function(d) {return d.flagged ? 'green' : 'red'},
                                stroke : '#222',
                                'stroke-width' : 2
                            })
            }

            scope.api.clear = function() {
                areaLayer.selectAll('*').remove();
                clearedareaLayer.selectAll('*').remove();
                itemLayer.selectAll('*').remove();
                nodeLayer.selectAll('*').remove();
                bombLayer.selectAll('*').remove();
            }

        } // /link
    };
}])
;
